import os
import importlib.util
import logging
from pathlib import Path

if (_config_home := os.path.expandvars('$XDG_CONFIG_HOME')) != '$XDG_CONFIG_HOME':
	CONFIG_HOME = Path(_config_home)
else:
	CONFIG_HOME = Path.home() / '.config'

DATA_PATH = CONFIG_HOME / 'konfsave'
CONFIG_FILENAME = 'konfsave.ini'
DEFAULT_CONFIG_PATH = Path(__file__).parent / 'default_config.ini'
FEATURES = {
	'GIT': True
}
FEATURE_REQUIREMENTS = {
	'GIT': ('pygit2',)
}


def feature_required(feature):
	def decorator(fn):
		if FEATURES.get(feature, None):
			return fn
		description = getattr(fn, '__name__', str(fn))
		return lambda *_, **__: logging.getLogger('konfsave').error(
			f'The feature "{feature}" is required to use this functionality ({description}). '
			f'Try installing it with pip: "pip install konfsave[{feature}]"'
		)
	return decorator


for k, v in FEATURE_REQUIREMENTS.items():
	for package in v:
		if importlib.util.find_spec(package) is None:
			FEATURES[k] = False
			break



import sys
import argparse
import itertools
import json
import logging
import subprocess
import time
import zipfile
from pathlib import Path
from typing import List

from . import constants
from . import config
from . import profiles

_N_T = '\n  '  # Backslashes are not allowed in f-string expressions, so use a variable
HELP_TEXT = '''Konfsave is a KDE config manager.

usage: konfsave <action> [args ...] [flags ...]

Actions:
help, --help, -h    print this message and exit
i, info, ls         get info about the current configuration, or a profile if specified
s, save             save the current configuration
l, load             load a saved profile
c, change           modify a profile's attributes
a, archive          export a profile as a ZIP file
u, unarchive        import an archived profile
f, files            list files that save would copy
g, groups           list default or available file groups

To see detailed usage instructions, run `konfsave <action> --help`.
All flags starting with '--' can be abbreviated.

To learn more about how to configure Konfsave, visit the GitHub wiki at
  https://github.com/selplacei/konfsave/wiki
or download it locally:
  git clone https://github.com/selplacei/konfsave.wiki.git
'''
logger = logging.getLogger('konfsave')


def parse_arguments(argv):
	# Actions need to parsed separately because each action has its own
	# argument format, so different ArgumentParser objects have to be used.
	if argv[0] == 'python':
		argv = argv[1:]
	action = argv[1] if len(argv) > 1 else 'help'
	try:
		next(v for k, v in {
			('-h', '--help', 'help'): lambda *_: print(HELP_TEXT),
			('i', 'info', 'ls'): action_info,
			('f', 'files'): action_list_files,
			('g', 'groups'): action_list_groups,
			('s', 'save'): action_save,
			('l', 'load'): action_load,
			('c', 'change'): action_change,
			('d', 'delete'): action_delete,
			('a', 'archive'): action_archive,
			('u', 'unarchive'): action_unarchive
		}.items() if action in k)(argv[2:])
	except StopIteration:
		logger.error(f'Unrecognized action: {action}\nTry \'konfsave help\' for more info.\n')
	except KeyboardInterrupt:
		print('Action cancelled.')


def action_info(argv):
	parser = argparse.ArgumentParser(prog='konfsave info')
	parser.add_argument(
		'profile', nargs='?', metavar='profile_name',
		help='If provided, prints detailed information about the profile.'
	)
	args = parser.parse_args(argv)
	if profile := args.profile:
		profiles.validate_profile_name(profile)
		info = profiles.profile_info(profile)
		if info is None:
			print(f'The profile {profile} doesn\'t exist.')
		else:
			print(f'Name: {info["name"]}')
			print(f'Stored at: {config.profile_home / profile}')
			print(f'Author: {info["author"] or "Unknown"}')
			print(f'Supported groups: {info["groups"] or "(unspecified)"}')
			if description := info['description']:
				print(description)
			else:
				print('No description provided.')
	else:
		if current_profile := profiles.current_profile():
			print(f'Current profile: {current_profile}')
		else:
			print(f'No profile is currently active.')
		# Choose directories which are valid profile names and contain an info file;
		# then, convert paths to names, and sort the final list of profiles.
		if saved_profiles := sorted(filter(
			lambda n: profiles.validate_profile_name(n, False),
			map(
				lambda q: q.name,
				filter(
					lambda p: (p / config.profile_info_filename).exists(),
					config.profile_home.glob('*')
				)
			)
		), key=str.lower):
			print(f'Saved profiles:\n  {_N_T.join(saved_profiles)}')
		else:
			print('No profiles are saved.')


def action_list_files(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave files',
		description='Print the list of files that are saved and loaded by default.'
	)
	parser.add_argument(
		'--include', '-i', action='extend', nargs='*', metavar='FILE', default=[],
		help='Same as in `save` or `load`.'
	)
	parser.add_argument(
		'--exclude', '-e', action='extend', nargs='*', metavar='FILE', default=[],
		help='Same as in `save` or `load`.'
	)
	args = parser.parse_args(argv)
	try:
		print('\n'.join(sorted(map(str, profiles.paths_to_save(args.include, args.exclude)), key=str.lower)))
	except ValueError as e:
		print(str(e))


def action_list_groups(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave groups',
		description='Print the list of default or available file groups, as specified in the config.'
	)
	parser.add_argument(
		'--available', '-a', action='count', default=0,
		help='If supplied once, print all available metagroups instead of just those enabled by default. '
		'If supplied twice, print all available groups.'
	)
	parser.add_argument(
		'--verbose', '-v', action='count', default=0,
		help='List the contents of groups. If supplied twice, break down sub-groups into paths. '
		'If supplied thrice, also expand directories into the files they contain.'
	)
	parser.add_argument(
		'--json', '-j', action='store_true',
		help='Print the output as a JSON string.'
	)
	args = parser.parse_args(argv)
	if args.available >= 2:
		groups = config.paths.keys()
	elif args.available:
		groups = config.metagroups.keys()
	else:
		groups = config.save_list
	groups = sorted(groups, key=str.lower)
	contents = {}
	if args.verbose >= 3:
		for group in groups:
			contents[group] = set(map(str, itertools.chain.from_iterable(
				map(profiles.expand_path, profiles.resolve_group(group))
			)))
	elif args.verbose == 2:
		for group in groups:
			contents[group] = profiles.resolve_group(group)
	elif args.verbose == 1:
		for group in groups:
			contents[group] = config.definitions[group]
	# else: leave ``contents`` empty
	if args.json:
		data = {k: list(map(str, v)) for k, v in contents.items()} if contents else list(groups)
		print(json.dumps(data))
	else:
		if args.available >= 2:
			print('All available groups:')
		elif args.available == 1:
			print('Available metagroups:')
		else:
			print('Default groups:')
		if contents:
			for k, v in contents.items():
				print(k, end='')
				print(_N_T + _N_T.join(map(str, v)))
		else:
			print(', '.join(groups))


def action_save(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave save',
		# The default usage string puts "[name]" at the end, for some reason.
		usage='konfsave save [-h] [name] [--destination DEST] [--follow-symlinks] '
		'[--include [FILE ...]] [--exclude [FILE ...]]'
	)
	parser.add_argument(
		'profile', metavar='name', nargs='?', default=profiles.current_profile(),
		help='Save as the specified profile instead of using the currently loaded name. '
		'This is required if no profile is active.'
	)
	parser.add_argument(
		'--destination', '-d', metavar='DEST', type=Path,
		help='Instead of saving to Konfsave\'s profile storage, save to a specified destination. '
	)
	parser.add_argument(
		'--follow-symlinks', '-s', action='store_true', dest='follow_symlinks',
		help='By default, symlinks are copied as symlinks. If this flag is used, '
		'the contents of symlinked files will be copied.'
	)
	parser.add_argument(
		'--include', '-i', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files or groups to add to the profile. Paths must be either absolute or relative '
		'to the home directory. Group names must start with a colon (:). Available groups and '
		'those loaded by default can be checked using `konfsave groups`.'
	)
	parser.add_argument(
		'--exclude', '-e', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files or groups to exclude from the profile; that is, they will not be copied, '
		'but if the files already exist in the profile, they will not be deleted. '
		'The format is the same as for --include. '
	)
	args = parser.parse_args(argv)
	if args.profile:
		profiles.validate_profile_name(args.profile)
	if args.profile \
		and (config.profile_home / args.profile).exists() \
		and (current := profiles.profile_info()) \
		and current['name'] != args.profile:
		if input(
			f'Warning: the profile "{args.profile}" already exists.'
			'Are you sure you want to overwrite it? [y/N]: '
		) != 'y':
			return
	include = set()
	exclude = set()
	for path in map(Path, args.exclude):
		if not path.is_absolute():
			path = Path.home() / path
		exclude.add(path)
	for path in map(Path, args.include):
		if not path.is_absolute():
			path = Path.home() / path
		include.add(path)
	profiles.save(
		name=args.profile,
		destination=args.destination,
		follow_symlinks=args.follow_symlinks,
		include=include,
		exclude=exclude
	)
	print('Success')


def action_load(argv):
	parser = argparse.ArgumentParser(prog='konfsave load')
	parser.add_argument(
		'profile',
		help='The name of the profile to load.'
	)
	parser.add_argument(
		'--overwrite',
		help='By default, loading will fail if no profile is active (i.e. the current configuration '
		'is not saved). Otherwise, the user will be asked for confirmation. '
		'Using --overwrite will bypass both of these checks.'
	)
	parser.add_argument(
		'--no-restart', '-n', action='store_false', dest='restart',
		help='After loading a profile, the Plasma shell is restarted unless this flag is specified.'
	)
	parser.add_argument(
		'--include', '-i', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files or groups to load from the profile in addition to those loaded by default. '
		'Paths must point to their final destination and be either absolute '
		'or relative to the home directory. '
		'Files that are loaded by default can be checked using `konfsave list-files`. '
		'Group names must start with a colon (:). '
		'Available groups and those loaded by default can be checked using `konfsave groups`.'
	)
	parser.add_argument(
		'--exclude', '-e', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files or groups to not load from the profile. This overrides default configuration. '
		'The format is the same as for --include.'
	)
	args = parser.parse_args(argv)
	profiles.validate_profile_name(args.profile)
	include = set()
	exclude = set()
	for path in map(Path, args.exclude):
		if not path.is_absolute():
			path = Path.home() / path
		exclude.add(path)
	for path in map(Path, args.include):
		if not path.is_absolute():
			path = Path.home() / path
		include.add(path)
	success = not profiles.load(
		args.profile,
		include,
		exclude,
		overwrite_unsaved_configuration=args.overwrite,
		restart=args.restart
	)
	if success:
		print('Success')


def action_change(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave change'
	)
	parser.add_argument(
		'profile', nargs='?',
		help='Profile to change. If not specified, the current profile will be used. '
		'Note that this will change the profile\'s saved attributes as well; '
		'if you\'d like to store new attributes as a separate profile, '
		'save the configuration under a new name first.'
	)
	parser.add_argument(
		'--name', '-n', help='New name for the profile.'
		'Changing this will also rename the profile\'s directory.'
	)
	parser.add_argument(
		'--author',
		help='New author.'
	)
	parser.add_argument(
		'--description',
		help='New description.'
	)
	parser.add_argument(
		'--groups', action='extend', nargs='*', metavar='NAME', default=[],
		help='New set of groups that this profile supports (don\'t include colons).'
	)
	args = parser.parse_args(argv)
	results = {k: v for k, v in vars(args).items() if k != 'profile' and v}
	if results:
		profiles.change(results, args.profile)
		print('Success')
	else:
		parser.print_help()


def action_delete(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave delete',
		description='Delete profile(s). Current system configuration will be unchanged; '
		'however, if a deleted profile is active, there will no longer be an active profile.'
	)
	parser.add_argument(
		'profile', action='extend', nargs='*', help='The profile(s) to delete.'
	)
	parser.add_argument('--noconfirm', action='store_false', dest='confirm')
	args = parser.parse_args(argv)
	if args.profile:
		success = not profiles.delete(args.profile, confirm=args.confirm)
		if success:
			print('Done')
	else:
		parser.print_help()


def action_archive(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave archive', description='Export/archive a profile to share or import later.',
		# The default usage puts "profile" at the end
		usage='konfsave archive [-h] [profile] [--destination PATH] [--overwrite] '
		'[--compresslevel LEVEL] [--compression {store,lzma,deflate,bzip2}]'
	)
	parser.add_argument(
		'profile', nargs='?', help='The profile to archive.', default=profiles.current_profile()
	)
	parser.add_argument(
		'--destination', '-d', metavar='PATH', type=Path,
		help='The full path (including the filename) of the resulting archive. '
		f'By default, archives are saved as "[profile name].konfsave.zip" under the home directory.'
	)
	parser.add_argument(
		'--overwrite', '-o', action='store_true',
		help='Unless this option is specified, if the resulting file '
		'already exists, archiving will fail.'
	)
	parser.add_argument(
		'--compresslevel', type=int, default='9', metavar='LEVEL',
		help='How much to compress the archive, from 1 to 9 (9 by default).'
		'This has no effect if --compression is "store" or "lzma".'
	)
	parser.add_argument(
		'--compression', choices=['store', 'lzma', 'deflate', 'bzip2'], default='bzip2',
		help='Compression method (bzip2 by default). "store" means no compression.'
	)
	args = parser.parse_args(argv)
	compression = {
		'store': zipfile.ZIP_STORED,
		'lzma': zipfile.ZIP_LZMA,
		'deflate': zipfile.ZIP_DEFLATED,
		'bzip2': zipfile.ZIP_BZIP2
	}[args.compression]
	try:
		profiles.archive_profile(
			profile=args.profile,
			overwrite=args.overwrite,
			destination=args.destination,
			compresslevel=args.compresslevel,
			compression=compression
		)
	except FileExistsError as e:
		logger.error(f'The file {e.filename} already exists.\n')
	except RuntimeError as e:
		logger.error(f'Error: {str(e)}\n')
	else:
		return
	logger.critical(f'Archiving "{args.profile}" failed.')


def action_unarchive(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave unarchive',
		description='Unpack and save a profile that was previously archived.',
		# The default usage puts "file" at the end
		usage='konfsave unarchive [-h] file [--name NAME] [--overwrite]'
	)
	parser.add_argument(
		'file', type=Path,
		help='Path to the archive to extract from.'
	)
	parser.add_argument(
		'--name', '-n', help='Extract to a specified profile name '
		'(by default, it\'s the same as the archived profile).'
	)
	parser.add_argument(
		'--overwrite', '-o', action='store_true',
		help='By default, if the destination profile already exists, '
		'the user will be asked for confirmation. '
		'Using this option will silently overwrite such profiles instead.'
	)
	args = parser.parse_args(argv)
	if args.name:
		profiles.validate_profile_name(args.name)
	if not profiles.unarchive_profile(
		source=args.file,
		new_name=args.name,
		overwrite=args.overwrite
	):
		print('Success')



import sys
import logging
import zipfile
import shutil
import json
from pathlib import Path

from konfsave import config
from konfsave import profiles


def archive_profile(profile, destination: Path = None, overwrite=False, compression=zipfile.ZIP_BZIP2, compresslevel=9):
	"""
	Archive a profile.
	
	If ``profile_dir`` isn't a valid profile, RuntimeError is raised.
	``destination`` should be a path pointing to the resulting .konfsave.zip file.
	If ``overwrite`` is False and the destination exists, ``FileExistsError`` will be raised.
	``compression`` and ``compresslevel`` are the same as in ``zipfile.ZipFile`` -
	see https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile for available values.
	"""
	open_mode = 'w' if overwrite else 'x'
	info = profiles.profile_info(profile)
	if info is None:
		raise RuntimeError(f'The directory {profile} is not a valid Konfsave profile.')
	profile_dir = config.profile_home / profile
	destination = destination or (config.archive_directory / (info['name'] + '.konfsave.zip'))
	with zipfile.ZipFile(destination, mode=open_mode, compression=compression, compresslevel=compresslevel) as zipf:
		print(f'Archiving "{profile}" into {destination}')
		for source in profile_dir.glob('**/*'):
			zipf.write(source, arcname=source.relative_to(profile_dir))
	print('Archiving finished')


def unarchive_profile(source: Path, new_name=None, overwrite=False, confirm=True) -> bool:
	"""
	Extract and import an archived profile without loading it.
	Returns True if unarchiving wasn't successful.
	If ``new_name`` is unspecified, the original archive's profile name is validated.
	``new_name`` is not validated in this function.
	
	If ``confirm`` is True, the user will be warned about the safety implications of unarchiving,
	and the function will exit if they cancel. Additionally, if ``confirm`` is True,
	``overwrite`` is False, and the destination exists, the user will be asked whether they want
	to overwrite the existing profile; if yes, the profile will be overwritten.
	If ``confirm`` and ``overwrite`` are False and the destination exists, FileExistsError will be raised.
	If ``new_name`` is specified, the profile will be loaded into the matching directory and
	its info will be updated.
	If the original archive has no information file and ``new_name`` is unspecified, ValueError will be raised.
	"""
	with zipfile.ZipFile(source) as zipf:
		with zipf.open(config.profile_info_filename) as infof:
			info = profiles.parse_profile_info(infof, convert_values=False)
			if info is None:
				profiles.logger.warning(
					f'The archive {source} has a malformed {config.profile_info_filename}.'
				)
				if new_name is None:
					raise ValueError(
						f'Could not infer the destination profile name for archive {source}. '
						'The archive doesn\'t contain a valid info file, and no new name was '
						'specified as a command line argument.'
					)
			if new_name:
				info['name'] = new_name
			else:
				profiles.validate_profile_name(info['name'])
		if confirm and input(
			'Warning: you\'re about to extract a profile that may have been created by someone else.\n'
			'Konfsave profiles can contain any file within the home directory, not just configurations.\n'
			'Unarchiving a profile will not load it; however, loading profiles from untrusted sources\n'
			'may have destructive consequences, including unintentionally overwriting personal data.\n'
			'Have you manually gone through the archive and made sure that every file is expected? [y/N]: '
		) != 'y':
			print('Unarchiving aborted.')
			return True
		destination = config.profile_home / info['name']
		backup = None
		if destination.exists() and not overwrite:
			if confirm:
				if input(
					f'Warning: the profile "{info["name"]}" is already saved.\n'
					'Are you sure you want to overwrite it? [y/N]: '
				) != 'y':
					print('Unarchiving aborted.')
					return True
				else:
					# Create a backup, which will be deleted if all of the next steps are successful.
					backup = Path(str(destination) + '.bkp')
					if backup.exists():
						profiles.logger.warning(
							f'Warning: the backup {backup} already exists. It will be overwritten.'
						)
						shutil.rmtree(backup)  # Path.rename() fails if the directory is not empty
					destination.rename(str(destination) + '.bkp')
			else:
				raise FileExistsError(filename=str(destination))
		try:
			zipf.extractall(
				destination,
				members=(p for p in zipf.namelist() if p != config.profile_info_filename)
			)
			with open(destination / config.profile_info_filename, 'w') as f:
				f.write(json.dumps(info))  # Write only after JSON serialization is successful
		except Exception:
			profiles.logger.exception(f'Unarchiving failed.\n')
			if backup:
				profiles.logger.warning(
					f'The previous version of "{info["name"]}" was backed up to {destination}'
				)
		else:
			if backup:
				shutil.rmtree(backup) 



#!/usr/bin/env python
import sys
import logging

from konfsave import constants
from konfsave import config
from konfsave import actions


def main():
	logging.basicConfig(format='[%(levelname)s] %(message)s')
	config.load_config()
	_constants = {k: str(v) for k, v in constants.__dict__.items() if k.isupper()}
	logging.getLogger('konfsave').debug(f'Constants: {_constants}')
	actions.parse_arguments(sys.argv)
	

if __name__ == '__main__':
	main()



import sys
import configparser
import itertools
import logging
import sys
from pathlib import Path
from typing import Set, Dict, List, Tuple

from . import constants

# The values referred to as "group names" include the preceding colon.

# Mapping of group names to what they contain, exactly as specified in the config
definitions: Dict[str, Set[str]] = {}
# Same as ``definitions``, but only contains metagroups (including redefined groups)
metagroups: Dict[str, Set[str]] = {}
# Mapping of group names to the paths they contain, with sub-groups recursively broken down
# Paths are absolute and resolved
paths: Dict[str, Set[Path]] = {}
# Set of files that should never be copied unless --included in the command line
exceptions = set()
# Default list of group names to save, as stored in [Defaults] -> save-list
save_list = []
profile_home: Path = None
profile_info_filename: str = None
current_profile_path: Path = None
archive_directory: Path = None


def default_paths() -> Tuple[Path]:
	"""
	Convert and return ``defaults['save-list']`` as a tuple of absolute and resolved ``Path``s.
	"""
	return tuple(itertools.chain.from_iterable(map(lambda g: paths[g], save_list)))


def load_config():
	global definitions, metagroups, paths, exceptions, save_list, profile_home
	global profile_info_filename, current_profile_path, archive_directory
	# Create the config file if missing
	if not (constants.DATA_PATH / 'konfsave.ini').exists():
		logging.getLogger('konfsave').warning('Config file missing, copying from default')
		constants.DATA_PATH.mkdir(parents=True, exist_ok=True)
		with open(constants.DATA_PATH / 'konfsave.ini', 'w') as f, open(constants.DEFAULT_CONFIG_PATH) as d:
			f.write(d.read())

	# Load the config file
	config = configparser.ConfigParser(allow_no_value=True, interpolation=_SpecialExtendedInterpolation())
	config.optionxform = str
	with open(constants.DATA_PATH / 'konfsave.ini') as f:
		config.read_file(f)
	
	# Set the logging level
	loglevel = {
		'DEBUG': logging.DEBUG,
		'INFO': logging.INFO,
		'WARNING': logging.WARNING,
		'ERROR': logging.ERROR,
		'CRITICAL': logging.CRITICAL
	}[config['Defaults']['log-level'].upper()]
	logging.basicConfig(level=loglevel)
	
	# Load exceptions
	for path in itertools.chain(
		map(
			lambda v: (Path.home() / Path(v)).resolve(),
			config['Home Directory Exceptions'].keys()
		),
		map(
			lambda v: (constants.CONFIG_HOME / Path(v)).resolve(),
			config['XDG_CONFIG_HOME Exceptions'].keys()
		)
	):
		exceptions.add(path)

	# Load path definitions
	for path, groups in itertools.chain(
		map(
			lambda v: ((Path.home() / Path(v[0])).resolve(), v[1]),
			config['Home Directory Path Definitions'].items()
		),
		map(
			lambda v: ((constants.CONFIG_HOME / Path(v[0])).resolve(), v[1]),
			config['XDG_CONFIG_HOME Path Definitions'].items()
		)
	):
		for group in groups.split(','):
			definitions.setdefault(f':{group}', set()).add(path)
	
	# Load metagroups
	for metagroup, subgroups in config['Metagroup Definitions'].items():
		if subgroups is None:
			# Redefenition of a normal group as a metagroup
			try:
				metagroups[f':{metagroup}'] = definitions[f':{metagroup}']
			except KeyError as e:
				raise ValueError(
					f'Attempted to redefine "{metagroup}" as a metagroup, but that group doesn\'t exist. '
					'Perhaps you forgot to include some files into it?'
				) from e
		else:
			# Creation of a new metagroup
			subgroups = set(map(lambda s: f':{s}', subgroups.split(',')))
			definitions[f':{metagroup}'] = subgroups
			metagroups[f':{metagroup}'] = subgroups
	
	# Recursively convert groups into paths
	undefined_groups = set()
	for group, definition in definitions.items():
		paths[group] = set()
		if isinstance(definition, Path):
			paths[group].add(definition)
			continue
		subvalues = list(definitions[group])
		while subvalues:
			subvalue = subvalues.pop()
			if isinstance(subvalue, Path):
				paths[group].add(subvalue)
			else:
				try:
					subvalues += list(definitions[subvalue])
				except KeyError:
					undefined_groups.add(subvalue)
	
	if undefined_groups:
		logging.getLogger('konfsave').info(
			f'The following groups are referenced in metagroup definitions, but are not defined: '
			+ ', '.join(sorted(undefined_groups))
		)
	
	# Load defaults
	save_list = list(map(lambda s: f':{s}', config['Defaults']['save-list'].split(',')))
	try:
		profile_home = Path(config['Defaults']['profile-home'])
		profile_info_filename = config['Defaults']['profile-info-filename']
		current_profile_path = Path(config['Defaults']['current-profile-path'])
		archive_directory = Path(config['Defaults']['archive-directory'])
	except KeyError:
		logging.getLogger('konfsave').critical(
			'Important values are missing from the config file. Did you recently update Konfsave?\n'
			'The following keys are expected in the Defaults section:\n'
			'profile-home, profile-info-filename, current-profile-path, archive-directory\n'
			f'See {str(constants.DEFAULT_CONFIG_PATH)} for example configuration,\nor simply delete '
			f'{str(constants.DATA_PATH / constants.CONFIG_FILENAME)} to reset the configuration completely.'
		)
		sys.exit(1)


class _SpecialExtendedInterpolation(configparser.ExtendedInterpolation):
	"""
	Identical to ``ExtendedInterpolation``, but also recognizes the following values:
		HOME
		CONFIG_HOME
		DATA_PATH
	which are identical to those stored in the ``constants`` module.
	"""
	def before_get(self, parser, section, option, value, defaults):
		return super().before_get(
			parser, section, option,
			value.replace(
				'${HOME}', str(Path.home())
			).replace(
				'${CONFIG_HOME}', str(constants.CONFIG_HOME)
			).replace(
				'${DATA_PATH}', str(constants.DATA_PATH)
			),
			defaults
		)



import itertools
import json
import logging
import os
import shutil
from pathlib import Path
from typing import Optional, Set, Union, TextIO, Iterable

from konfsave import config
from konfsave import constants
from konfsave import profiles


def validate_profile_name(name, exit_if_invalid=True) -> bool:
	valid = name.isidentifier()
	if (not valid) and exit_if_invalid:
		profiles.logger.critical(f'The profile name "{name}" is invalid.')
		sys.exit(1)
	else:
		profiles.logger.debug(f'The profile name "{name}" is invalid.')
	return valid


def copy_allow_samefile(*args, **kwargs):
	"""
	Identical to ``shutil.copy()``, but doesn't raise an exception when copying the same file.
	"""
	try:
		shutil.copy(*args, **kwargs)
	except shutil.SameFileError:
		pass


def copy_path(source, destination, overwrite=True, follow_symlinks=False):
	profiles.logger.info(f'Copying {source}')
	destination.parent.mkdir(parents=True, exist_ok=True)
	if source.is_dir():
		shutil.copytree(
			src=source,
			dst=destination,
			symlinks=follow_symlinks,
			copy_function=copy_allow_samefile,
			dirs_exist_ok=overwrite
		)
	else:
		if overwrite or not destination.exists():
			copy_allow_samefile(source, destination, follow_symlinks=follow_symlinks)
			
			
def current_profile():
	try:
		return profile_info()['name']
	except TypeError:
		return None


def resolve_group(val) -> Set[Path]:
	"""
	Resolve a possible group, as specified in the user's config.
	If ``val`` is os.PathLike, it's converted to Path, resolved, and returned as a single item in a set.
	``val`` is considered a group if it starts with a colon.
	If the group isn't defined, KeyError is raised.
	If ``val`` doesn't either represent a path or start with a colon, ValueError is raised.
	"""
	if isinstance(val, os.PathLike):
		return {Path(val).resolve()}
	elif isinstance(val, str) and val.startswith(':'):
		return config.paths[val]  # Raises KeyError if the group is undefined
	else:
		raise ValueError(f'The value "{val}" is not a path or a group name.')


def expand_path(path) -> Set[Path]:
	"""
	If ``path`` points to a directory, return all files within the directory (recursively).
	Otherwise, return a set that contains ``path`` as its sole member.
	"""
	if path.is_dir():
		return set(path.glob('**/*'))
	return {path}


def paths_to_save(include=None, exclude=None, default_include=None) -> Set[Path]:
	"""
	Calculate and return a set of files to save to or load from a profile.
	Paths are returned as absolute and resolved,
	and point to the actual files in the home directory (never to directories).
	Directories specified in ``include``, ``exclude``, and
	``default_include`` are traversed recursively.
	
	The optional parameters ``include`` and ``exclude`` represent overrides, typically given by
	the user as command line arguments. They will always take priority over other configuration.
	The optional parameter ``default_include`` represents a list of default paths to include,
	and is typically not specified. If it is None, the list is read from the config's save-list.
	If it is an empty iterator, only values specified in ``include`` and ``exclude`` are considered.
	
	``include``, ``exclude``, and ``default_include`` must be given either as
	absolute paths (os.PathLike) or groups (starting with a colon).
	"""
	include = {*itertools.chain.from_iterable(map(resolve_group, include or ()))}
	include = {*itertools.chain.from_iterable(map(expand_path, include))}
	exclude = {*itertools.chain.from_iterable(map(resolve_group, exclude or ()))}
	exclude = {*itertools.chain.from_iterable(map(expand_path, exclude))}
	for exception in itertools.chain.from_iterable(map(expand_path, config.exceptions)):
		if exception not in include:
			exclude.add(exception)
	if default_include:
		default_include = {*itertools.chain.from_iterable(map(resolve_group, default_include))}
	else:
		default_include = {*itertools.chain.from_iterable(map(resolve_group, config.default_paths()))}
	default_include = {*itertools.chain.from_iterable(map(expand_path, default_include))}
	return (default_include | include) - exclude


_profile_info_cache = {}

def profile_info(profile_name=None, convert_values=True, use_cache=True) -> Optional[dict]:
	"""
	If the profile name is invalid, this function will print a warning and continue normally.
	
	When no argument is supplied, this will read ``config.current_profile_path``.
	The return value is ``None`` if the JSON file is missing or malformed.
	"""
	if use_cache and profile_name in _profile_info_cache:
		return _profile_info_cache[profile_name]
	if profile_name and not validate_profile_name(profile_name, exit_if_invalid=False):
		profiles.logger.warning(f'"f{profile_info}" is an invalid profile name\n')
	try:
		info = parse_profile_info(
			(config.profile_home / profile_name / config.profile_info_filename) \
				if profile_name else config.current_profile_path, convert_values=convert_values
		)
		if use_cache and info:
			_profile_info_cache[profile_name] = info
			return info.copy()
		return info
	except FileNotFoundError:
		return None


def parse_profile_info(profile_info_file: Union[TextIO, os.PathLike], convert_values=True) -> Optional[dict]:
	try:
		close_later = False
		if isinstance(profile_info_file, os.PathLike):
			f = open(profile_info_file)
			close_later = True
		else:
			f = profile_info_file
		info = json.load(f)
		if close_later:
			f.close()
		assert info['name'].isidentifier()
		info['author'] = info.get('author', None)
		info['description'] = info.get('description', None)
		info['groups'] = info.get('groups', None)
		# Currently, there are no values to convert.
		return info
	except (json.JSONDecodeError, KeyError, AssertionError) as e:
		if isinstance(profile_info_file, os.PathLike):
			profiles.logger.warning(f'Malformed profile info at {profile_info_file}\n{str(e)} \n')
		return None 



import shutil
import subprocess
import time
from pathlib import Path

from konfsave import constants
from konfsave import config
from konfsave import profiles


def load(name, include=None, exclude=None, overwrite_unsaved_configuration=False, restart=True) -> bool:
	"""
	The name is not validated in this function.
	True is returned if the user canceled the action.
	
	The KDE configuration will be overwritten if:
		* ``overwrite_unsaved_configuration`` is True
			OR all of the following is true:
		* A profile is active
		* The source profile's info JSON is valid
		* The user manually confirmed that they want to overwrite their configuration
	"""
	profile_root = config.profile_home / name
	if overwrite_unsaved_configuration is not True:  # Be really sure that overwriting is intentional
		# If the checks below fail, exit the function.
		try:
			current_info = profiles.profile_info()
			assert (profile_root / config.profile_info_filename).is_file()
			if current_info is None:
				print(
					'Warning: there is no active profile, and the current configuration is NOT SAVED.'
				)
			elif current_info['name'] != name:
				print(
					'Warning: you\'re loading a new profile. Konfsave cannot check if the currently '
					'active profile has been updated with current files.'
				)
			else:
				print(
					'Warning: you\'re overwriting the current profile with existing configuration. '
					'Konfsave cannot check if the saved files are up-to-date.'
				)
			if input(
				'Are you sure you want to overwrite the current system configuration? [y/N]: '
			).lower() != 'y':
				print('Loading aborted.')
				return True
		except Exception as e:
			profiles.logger.error('Refusing to overwrite unsaved configuration due to an error')
			raise
	if restart:
		restart_list = []
		subprocess.run(
			['kquitapp5', 'plasmashell'],
			stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
		)
		try:
			# Check if Latte Dock is running
			subprocess.run(
				['ps', '-C', 'latte-dock'],
				check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
			)
		except subprocess.CalledProcessError:
			profiles.logger.info('No running instance of Latte detected')
		else:
			subprocess.run(['kquitapp5', 'lattedock'])
			restart_list.append('latte-dock')
	config.current_profile_path.unlink(missing_ok=True)
	for path in map(
		lambda p: Path(p).relative_to(Path.home()),
		profiles.paths_to_save(include, exclude)
	):
		source = profile_root / path
		if source.exists():
			profiles.copy_path(source, Path.home() / path)
		else:
			profiles.logger.info(f'The file {source} doesn\'t exist. Skipping\n')
	shutil.copyfile(profile_root / config.profile_info_filename, config.current_profile_path)
	if restart:
		subprocess.run(
			['kstart5', 'plasmashell'],
			stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
		)
		try:
			# Check if Kwin is running
			subprocess.run(
				['ps', '-C', 'kwin_x11'],
				check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
			)
		except subprocess.CalledProcessError:
			profiles.logger.info('No running instance of KWin detected')
		else:
			# If so, reload Kwin
			subprocess.run(
				['dbus-send', '--session', '--dest=org.kde.KWin', '/KWin', 'org.kde.KWin.reloadConfig'],
				stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
			)
		if 'latte-dock' in restart_list:
			time.sleep(3)  # Allow KWin to completely restart
			subprocess.run(
				['kstart5', 'latte-dock'],
				stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
			) 



import json
import shutil
from pathlib import Path
from typing import Union, Iterable

from konfsave import config
from konfsave import profiles


def change(results, profile=None):
	"""
	``results`` is a dictionary containing modified attributes and their values.
	Attributes that exist in the profile but are not present in ``results`` will be kept as is.
	
	If the name is changed, ``rename()`` will be called. Additionally, if ``profile`` is
	the same as the current profile, the current profile's attributes will be changed as well.
	"""
	current = profiles.current_profile()
	if profile is None:
		if current is None:
			raise RuntimeError(
				'Attempted to change the current profile, but no profile is active. Try saving or '
				'loading something first, or specifying which saved profile you\'d like to change.'
			)
		profile = current
	new_info = profiles.profile_info(profile, convert_values=False).copy()
	new_info.update(results)
	if profile == current:
		# Also modify the profile info stored in the home directory
		with open(config.current_profile_path, 'w') as f:
			f.write(json.dumps(new_info))
	if 'name' in results:
		rename(profile, results['name'], change_info=False)  # Avoid writing to the file twice
	with open(config.profile_home / new_info['name'] / config.profile_info_filename, 'w') as f:
		f.write(json.dumps(new_info))  # Write only after JSON serialization is successful
		
		
def rename(source, result, change_info=True):
	"""
	Rename a saved profile by changing its info and directory. Only the resulting name is validated.
	The current profile is not modified even if its name matches the source name.
	"""
	profiles.validate_profile_name(result)
	if (config.profile_home / result).exists():
		raise FileExistsError(f'A profile named "{result}" is already saved.')
	if not (config.profile_home / source).exists():
		raise RuntimeError(f'The profile "{source}" doesn\'t exist.')
	if change_info:
		info = profiles.profile_info(source)
		info.update({'name': result})
		with open(config.profile_home / source / config.profile_info_filename, 'w') as f:
			f.write(json.dumps(new_info))  # Write only after JSON serialization is successful
	(config.profile_home / source).rename(config.profile_home / result)


def delete(profile: Union[str, Iterable[str]], clear_active=True, confirm=True) -> bool:
	"""
	``profile`` may be a string specifying a single profile to delete, or a list of profiles.
	
	If ``clear_active`` is True and the deleted profile has the same name as the active profile,
	the active profile info will be deleted as well. Current configuration will be unaffected.
	
	True is returned if the user canceled the action, or if it is otherwise unsuccessful.
	"""
	if not isinstance(profile, str):
		profile = list(profile)
		if len(profile) > 1:
			_profile_list = ', '.join(map(lambda n: f'"{n}"', profile))
			if confirm:
				print(f'Warning: you\'re about to delete the profiles {_profile_list}.')
				if input('Are you sure you want to permanently delete all of them? [y/N]: ') != 'y':
					print('Deleting aborted.')
					return True
			for prf in profile:
				delete(prf, clear_active=clear_active, confirm=False)
			return
		elif len(profile) == 1:
			profile = profile[0]
		else:
			profiles.logger.error('The list of profiles to delete is empty.')
			return True
	if not (config.profile_home / profile).exists():
		profiles.logger.error(f'The profile "{profile}" doesn\'t exist.')
		return True
	if confirm:
		print(f'Warning: you\'re about to delete the profile "{profile}".')
		if input('Are you sure you want to permanently delete it? [y/N]: ') != 'y':
			print('Deleting aborted.')
			return True
	if clear_active and profile == profiles.current_profile():
		config.current_profile_path.unlink(missing_ok=True)
	shutil.rmtree(config.profile_home / profile)
	print(f'Deleted profile "{profile}"')



import json
from pathlib import Path

from konfsave import config
from konfsave import profiles


def save(name=None, include=None, exclude=None, follow_symlinks=False, destination=None):
	"""
	The name is not validated in this function.
	
	If ``name`` is unspecified, the current profile's name is used.
	Otherwise, the current profile will be switched to the result.
	``include`` and ``exclude`` must be given in the same format as to ``paths_to_save()``.
	"""
	info = profiles.profile_info(name, convert_values=False)
	if name is None:
		if info is None:
			raise RuntimeError(
				'Attempted to save the current profile, but no profile is active. '
				'Perhaps you\'re saving new configuration but forgot to specify a profile name?'
			)
		else:
			name = info['name']
	profile_dir = (config.profile_home / name) if destination is None else destination
	profile_dir.mkdir(parents=True, exist_ok=True)
	for path in map(Path, profiles.paths_to_save(include, exclude)):
		if not path.exists():
			profiles.logger.info(f'The path {path} doesn\'t exist. Skipping')
		elif not path.is_relative_to(Path.home()):
			profiles.logger.warning(f'The path {path} is not within the user\'s home directory. Skipping')
		else:
			profiles.copy_path(path, profile_dir / path.relative_to(Path.home()), follow_symlinks=follow_symlinks)
	new_info = {
		'name': name,
		'author': info['author'] if info else None,
		'description': info['description'] if info else None,
		'groups': info['groups'] if info else []
	}
	with open(profile_dir / config.profile_info_filename, 'w') as f:
		f.write(json.dumps(new_info))  # Write only after JSON serialization is successful
	with open(config.current_profile_path, 'w') as f:
		f.write(json.dumps(new_info))  # Write only after JSON serialization is successful 
